/*
 *   Copyright 2011 Institute of Communication and Computer Systems (ICCS) 
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 */

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <dmmlib/heap.h>
#include "dmm_config.h"
#ifdef HAVE_LOCKS
#include "posix_lock.h"
#endif /* HAVE_LOCKS */
#include "other.h"
#include "sys_alloc.h"
#include "block_header.h"
#ifdef WITH_STATS
#include "print_stats.h"
#endif /* WITH_STATS */

void *sys_alloc(allocator_t *allocator, heap_t *heap, size_t size) {

    size_t allocation_size, previous_size, previous_size_availability;
    void *ptr;

#ifdef HAVE_LOCKS
    sbrk_lock();
#endif /* HAVE_LOCKS */

    allocation_size = req_padding(size) + HEADER_SIZE;

    /* Before the first application malloc() call, the allocator uses
     * sys_alloc() to allocate space for its metadata. So, these are the first
     * data blocks in the heaps and there are no previous data blocks.
     */
#ifndef WITH_MEMORY_SPACE_AWARENESS
    if(allocator->border_ptr != NULL) {
#else
    /* Note: border_ptr has already a value in memory space aware allocators,
     * so in this case we have to check also the initialized value.
     */
    if(allocator->border_ptr != NULL && allocator->initialized) {
#endif /* WITH_MEMORY_SPACE_AWARENESS */
        previous_size = get_size(allocator->border_ptr);
        previous_size_availability = get_size_availability(allocator->border_ptr);
    } else {
        previous_size = 0;
        previous_size_availability = 1; /* Occupied and of 0 size */
        allocator->initialized = true;
    }

#ifndef WITH_MEMORY_SPACE_AWARENESS
    ptr = sbrk((int) allocation_size);
    if(ptr == (void *) -1) {
        printf("sbrk problem for size of: %zu\n", allocation_size);
        printf("Error on sbrk: %s\n", strerror( errno ) );
        print_stats(allocator);
        return NULL;
    }   
#if defined (COALESCING_FIXED) || defined (COALESCING_VARIABLE)
    if(allocator->border_ptr != NULL && ptr != (void *) 
            ((char *) allocator->border_ptr + previous_size)) {
        printf("sbrk() does not return sequential space. "
                "You should disable coalescing.\n");
    }
#endif /* COALESCING_FIXED || COALESCING_VARIABLE */    
#else
    if(allocator->remaining_size >= allocation_size) {
        ptr = (void *) ((char *) allocator->border_ptr + previous_size);
        allocator->remaining_size -= allocation_size;
    } else {
        printf("No more free space.\n");
#ifdef WITH_STATS
        print_stats(allocator);
#endif /* WITH_STATS */
        return NULL;
    }
#endif /* WITH_MEMORY_SPACE_AWARENESS */

    /* Go to the data part of the block */
    ptr = (void *) ((char *) ptr + HEADER_SIZE);
    
    /* Set the new border pointer */
    allocator->border_ptr = ptr;

    /* Set some values for block header */
    set_size(ptr, req_padding(size));
    set_previous_size_availability(ptr, previous_size_availability);

    /* FIXME To be refactored - START */
    set_requested_size(ptr, size);
    mark_used(ptr);

    /* Update the used blocks list */
    push_block(ptr, heap->used_blocks_head);

#ifdef WITH_STATS
    /* Update statistics */
    heap->dmm_stats.mem_allocated += req_padding(size);
    heap->dmm_stats.mem_requested += size;
    heap->dmm_stats.live_objects += 1;
    heap->dmm_stats.num_malloc += 1;
#endif /* WITH_STATS */

    /* FIXME To be refactored - END */

#ifdef HAVE_LOCKS
    sbrk_unlock();
#endif /* HAVE_LOCKS */

    return  ptr;
}

